from abc import ABC, abstractmethod

from sole.ioc.interfaces import ICommand


class IPlugin(ABC):
    @abstractmethod
    def get_load_command(self) -> ICommand:
        ...

    @abstractmethod
    def get_unload_command(self) -> ICommand:
        ...



from typing import Any

from .exceptions import ResolveDependencyException
from ._exception_helper import raise_
from ._ioc_base import IOCBaseContainer


ioc_base_container = IOCBaseContainer(
    lambda key: raise_(ResolveDependencyException(f"Dependency {key} is missing"))
)


def resolve(key: str, *args) -> Any:
    try:
        return ioc_base_container["IoC.Resolve"](key, *args)
    except ResolveDependencyException as e:
        raise e
    except Exception as e:
        raise ResolveDependencyException(
            f"An unexpected exception occurred with key: {key} and args: {args}: {e}"
        )



from abc import ABC, abstractmethod
from typing import Any, Callable


class ICommand(ABC):
    @abstractmethod
    def __call__(self) -> None:
        ...


class IResolveDependencyStrategy(ABC):
    @abstractmethod
    def __call__(self, *args: Any) -> Any:
        ...


class IDependenciesContainer(ABC):
    @abstractmethod
    def __init__(self, not_found_strategy: Callable):
        ...

    @abstractmethod
    def __getitem__(self, key: str) -> IResolveDependencyStrategy:
        ...

    @abstractmethod
    def __setitem__(self, key: str, strategy: IResolveDependencyStrategy):
        ...



import unittest

from sturdy import resolve
from sturdy.ioc.exceptions import ResolveDependencyException


class TestBase(unittest.TestCase):
    def test_ioc_resolve_missing_key_fail(self):
        self.assertRaises(ResolveDependencyException, resolve, "MISSING_KEY")

    def test_ioc_register_and_ioc_resolve_success(self):
        resolve("IoC.Register", "App.add", lambda *args: args[0] + args[1])()
        self.assertEqual(12, resolve("App.add", 1, 11))

    def test_ioc_register_no_args_fail(self):
        self.assertRaises(ResolveDependencyException, resolve, "IoC.Register")

    def test_ioc_register_bad_arg_fail(self):
        resolve("IoC.Register", "App.dummy", 1)()
        self.assertRaises(ResolveDependencyException, resolve, "App.dummy")



from typing import Any, Callable

from .interfaces import IDependenciesContainer, IResolveDependencyStrategy, ICommand
from .exceptions import ResolveDependencyException


class IOCBaseContainer(IDependenciesContainer):
    def __init__(self, not_found_strategy: Callable):
        self.__not_found_strategy = not_found_strategy
        self.__store = {}

        self.__store["IoC.Resolve"] = IOCBaseResolveResolver(self)
        self.__store["IoC.Register"] = IOCBaseRegisterCommandResolver(self)
        self.__store["IoC.BaseContainer"] = lambda: self

    def __getitem__(self, key: str) -> IResolveDependencyStrategy:
        try:
            return self.__store[key]
        except KeyError:
            return self.__not_found_strategy(key)
        except ResolveDependencyException as e:
            raise e

    def __setitem__(self, key: str, strategy: IResolveDependencyStrategy):
        self.__store[key] = strategy


class IOCBaseResolveResolver(IResolveDependencyStrategy):
    def __init__(self, container: IDependenciesContainer):
        self.__container = container

    def __call__(self, *args: Any) -> IResolveDependencyStrategy:
        key = args[0]
        return self.__container[key](*args[1:])


class IOCBaseRegisterCommandResolver(IResolveDependencyStrategy):
    def __init__(self, container: IDependenciesContainer):
        self.__container = container

    def __call__(self, *args: Any) -> IResolveDependencyStrategy:
        try:
            key = args[0]
            strategy = args[1]
            # TODO: Throw ResolveDependencyException "agrs[0] must have type String, args[1] must have IResolveDependencyStrategy"
            return IOCBaseRegisterCommand(self.__container, key, strategy)
        except IndexError:
            raise ResolveDependencyException(
                "IoC.Register requires two args: key(str) and strategy(IResolveDependencyStrategy)"
            )


class IOCBaseRegisterCommand(ICommand):
    def __init__(
        self,
        scope: IDependenciesContainer,
        key: str,
        strategy: IResolveDependencyStrategy,
    ):
        self.container = scope
        self.key = key
        self.strategy = strategy

    def __call__(self) -> None:
        self.container[self.key] = self.strategy



"""
python setup.py test
python3 -m build
twine upload --repository testpypi dist/* # https://upload.pypi.org/legacy/
twine upload --repository pypi dist/* # https://test.pypi.org/legacy/
rm -rf dist sturdy.egg-info
"""

import os
from setuptools import setup, find_packages

import sturdy


here = os.path.dirname(__file__)

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="sturdy",
    description="Modular app server",
    maintainer="Ivan Sharun",
    maintainer_email="ivan@sha.run",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/omstu-amcs/sturdy",
    license="MIT",
    project_urls={
        "Source": "https://github.com/omstu-amcs/sturdy",
    },
    keywords=["Plugin", "IoC", "Inversion of control", "resolve"],
    version=sturdy.__version__,
    packages=find_packages(),
    test_suite="tests",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
)
